@@ -15,6 +15,7 @@
 import hashlib
 import logging
 import operator
+import os
 import time
 
 from cinderclient.v1 import client as cinder_client
@@ -25,6 +26,7 @@ from ironicclient import client as ironic_client
 from ironicclient import exceptions as ironic_exceptions
 from keystoneclient import auth as ksc_auth
 from keystoneclient import session as ksc_session
+from keystoneclient import client as keystone_client
 from novaclient import client as nova_client
 from novaclient import exceptions as nova_exceptions
 from novaclient.v2 import floating_ips
@@ -61,10 +63,10 @@ class OpenStackCloudTimeout(OpenStackCloudException):
     pass
 
 
-def openstack_clouds(config=None):
+def openstack_clouds(config=None, debug=False):
     if not config:
         config = os_client_config.OpenStackConfig()
-    return [OpenStackCloud(f.name, f.region, **f.config)
+    return [OpenStackCloud(cloud=f.name, debug=debug, **f.config)
             for f in config.get_all_clouds()]
 
 
@@ -72,14 +74,33 @@ def openstack_cloud(debug=False, **kwargs):
     cloud_config = os_client_config.OpenStackConfig().get_one_cloud(
         **kwargs)
     return OpenStackCloud(
-        cloud_config.name, cloud_config.region,
+        cloud=cloud_config.name,
         debug=debug, **cloud_config.config)
 
 
-def operator_cloud(**kwargs):
+def operator_cloud(debug=False, **kwargs):
     cloud_config = os_client_config.OpenStackConfig().get_one_cloud(**kwargs)
     return OperatorCloud(
-        cloud_config.name, cloud_config.region, **cloud_config.config)
+        cloud_config.name, debug=debug, **cloud_config.config)
+
+
+def _ssl_args(verify, cacert, cert, key):
+    if cacert:
+        if not os.path.exists(cacert):
+            raise OpenStackCloudException(
+                "CA Cert {0} does not exist".format(cacert))
+        verify = cacert
+
+    if cert:
+        if not os.path.exists(cert):
+            raise OpenStackCloudException(
+                "Client Cert {0} does not exist".format(cert))
+        if key:
+            if not os.path.exists(key):
+                raise OpenStackCloudException(
+                    "Client key {0} does not exist".format(key))
+            cert = (cert, key)
+    return (verify, cert)
 
 
 def _get_service_values(kwargs, service_key):
@@ -106,53 +127,89 @@ def _iterate_timeout(timeout, message):
 
 
 class OpenStackCloud(object):
+    """Represent a connection to an OpenStack Cloud.
 
-    def __init__(self, cloud, region='',
+    OpenStackCloud is the entry point for all cloud operations, regardless
+    of which OpenStack service those operations may ultimately come from.
+    The operations on an OpenStackCloud are resource oriented rather than
+    REST API operation oriented. For instance, one will request a Floating IP
+    and that Floating IP will be actualized either via neutron or via nova
+    depending on how this particular cloud has decided to arrange itself.
+
+    :param string name: The name of the cloud
+    :param dict auth: Dictionary containing authentication information.
+                      Depending on the value of auth_plugin, the contents
+                      of this dict can vary wildly.
+    :param string region_name: The region of the cloud that all operations
+                               should be performed against.
+                               (optional, default '')
+    :param string auth_plugin: The name of the keystone auth_plugin to be used
+    :param string endpoint_type: The type of endpoint to get for services
+                                 from the service catalog. Valid types are
+                                 `public` ,`internal` or `admin`. (optional,
+                                 defaults to `public`)
+    :param bool private: Whether to return or use private IPs by default for
+                         servers. (optional, defaults to False)
+    :param bool verify: The verification arguments to pass to requests. True
+                        tells requests to verify SSL requests, False to not
+                        verify. (optional, defaults to True)
+    :param string cacert: A path to a CA Cert bundle that can be used as part
+                          of verifying SSL requests. If this is set, verify
+                          is set to True. (optional)
+    :param string cert: A path to a client certificate to pass to requests.
+                        (optional)
+    :param string key: A path to a client key to pass to requests. (optional)
+    :param bool debug: Enable or disable debug logging (optional, defaults to
+                       False)
+    :param int cache_interval: How long to cache items fetched from the cloud.
+                               Value will be passed to dogpile.cache. None
+                               means do not cache at all.
+                               (optional, defaults to None)
+    """
+
+    def __init__(self, cloud, auth,
+                 region_name='',
                  auth_plugin='password',
-                 insecure=False, verify=None, cacert=None, cert=None, key=None,
+                 endpoint_type='public',
+                 private=False,
+                 verify=True, cacert=None, cert=None, key=None,
                  debug=False, cache_interval=None, **kwargs):
 
         self.name = cloud
-        self.region = region
-
+        self.auth = auth
+        self.region_name = region_name
         self.auth_plugin = auth_plugin
-        self.auth = kwargs.get('auth')
-
-        self.region_name = kwargs.get('region_name', region)
-        self._auth_token = kwargs.get('auth_token', None)
+        self.endpoint_type = endpoint_type
+        self.private = private
 
         self.service_types = _get_service_values(kwargs, 'service_type')
         self.service_names = _get_service_values(kwargs, 'service_name')
         self.endpoints = _get_service_values(kwargs, 'endpoint')
         self.api_versions = _get_service_values(kwargs, 'api_version')
 
-        self.endpoint_type = kwargs.get('endpoint_type', 'publicURL')
-        self.private = kwargs.get('private', False)
-
-        if verify is None:
-            if insecure:
-                verify = False
-            else:
-                verify = cacert or True
-        self.verify = verify
-
-        if cert and key:
-            cert = (cert, key)
-        self.cert = cert
+        (self.verify, self.cert) = _ssl_args(verify, cacert, cert, key)
 
+        # dogpile.cache.memory does not clear things out of the memory dict,
+        # so there is a possibility of memory bloat over time. By passing
+        # None a user will not get any in-memory cache and thus not have
+        # a large memory leak.
+        if cache_interval is None:
+            backend_name = 'dogpile.cache.null'
+        else:
+            backend_name = 'dogpile.cache.memory'
         self._cache = cache.make_region().configure(
-            'dogpile.cache.memory', expiration_time=cache_interval)
+            backend_name, expiration_time=cache_interval)
         self._container_cache = dict()
         self._file_hash_cache = dict()
 
-        self.debug = debug
-
         self._keystone_session = None
+        self._auth_token = None
 
         self._cinder_client = None
         self._glance_client = None
         self._glance_endpoint = None
         self._ironic_client = None
+        self._keystone_client = None
         self._neutron_client = None
         self._nova_client = None
         self._swift_client = None
@@ -160,7 +217,7 @@ class OpenStackCloud(object):
 
         self.log = logging.getLogger('shade')
         log_level = logging.INFO
-        if self.debug:
+        if debug:
             log_level = logging.DEBUG
         self.log.setLevel(log_level)
         self.log.addHandler(logging.StreamHandler())
@@ -231,6 +288,21 @@ class OpenStackCloud(object):
                     "Error authenticating to the keystone: %s " % e.message)
         return self._keystone_session
 
+    @property
+    def keystone_client(self):
+        if self._keystone_client is None:
+            try:
+                self._keystone_client = keystone_client.Client(
+                    session=self.keystone_session,
+                    auth_url=self.keystone_session.get_endpoint(
+                        interface=ksc_auth.AUTH_INTERFACE))
+            except Exception as e:
+                self.log.debug(
+                    "Couldn't construct keystone object", exc_info=True)
+                raise OpenStackCloudException(
+                    "Error constructing keystone client: %s" % e.message)
+        return self._keystone_client
+
     @property
     def service_catalog(self):
         return self.keystone_session.auth.get_access(
@@ -242,6 +314,136 @@ class OpenStackCloud(object):
             self._auth_token = self.keystone_session.get_token()
         return self._auth_token
 
+    @property
+    def project_cache(self):
+        @self._cache.cache_on_arguments()
+        def _project_cache():
+            return {project.id: project for project in
+                    self._project_manager.list()}
+        return _project_cache()
+
+    @property
+    def _project_manager(self):
+        # Keystone v2 calls this attribute tenants
+        # Keystone v3 calls it projects
+        # Yay for usable APIs!
+        return getattr(
+            self.keystone_client, 'projects', self.keystone_client.tenants)
+
+    def _get_project(self, name_or_id):
+        """Retrieve a project by name or id."""
+
+        # TODO(mordred): This, and other keystone operations, need to have
+        #                domain information passed in. When there is no
+        #                available domain information, we should default to
+        #                the currently scoped domain which we can requset from
+        #                the session.
+        for id, project in self.project_cache.items():
+            if name_or_id in (id, project.name):
+                return project
+        return None
+
+    def get_project(self, name_or_id):
+        """Retrieve a project by name or id."""
+        project = self._get_project(name_or_id)
+        if project:
+            return meta.obj_to_dict(project)
+        return None
+
+    def update_project(self, name_or_id, description=None, enabled=True):
+        try:
+            project = self._get_project(name_or_id)
+            return meta.obj_to_dict(
+                project.update(description=description, enabled=enabled))
+        except Exception as e:
+            self.log.debug("keystone update project issue", exc_info=True)
+            raise OpenStackCloudException(
+                "Error in updating project {project}: {message}".format(
+                    project=name_or_id, message=e.message))
+
+    def create_project(self, name, description=None, enabled=True):
+        """Create a project."""
+        try:
+            self._project_manager.create(
+                project_name=name, description=description, enabled=enabled)
+        except Exception as e:
+            self.log.debug("keystone create project issue", exc_info=True)
+            raise OpenStackCloudException(
+                "Error in creating project {project}: {message}".format(
+                    project=name, message=e.message))
+
+    def delete_project(self, name_or_id):
+        try:
+            project = self._get_project(name_or_id)
+            self._project_manager.delete(project.id)
+        except Exception as e:
+            self.log.debug("keystone delete project issue", exc_info=True)
+            raise OpenStackCloudException(
+                "Error in deleting project {project}: {message}".format(
+                    project=name_or_id, message=e.message))
+
+    @property
+    def user_cache(self):
+        @self._cache.cache_on_arguments()
+        def _user_cache():
+            return {user.id: user for user in
+                    self.keystone_client.users.list()}
+        return _user_cache()
+
+    def _get_user(self, name_or_id):
+        """Retrieve a user by name or id."""
+
+        for id, user in self.user_cache.items():
+            if name_or_id in (id, user.name):
+                return user
+        return None
+
+    def get_user(self, name_or_id):
+        """Retrieve a user by name or id."""
+        user = self._get_user(name_or_id)
+        if user:
+            return meta.obj_to_dict(user)
+        return None
+
+    def update_user(self, name_or_id, email=None, enabled=True):
+        try:
+            user = self._get_user(name_or_id)
+            return meta.obj_to_dict(
+                user.update(email=email, enabled=enabled))
+        except Exception as e:
+            self.log.debug("keystone update user issue", exc_info=True)
+            raise OpenStackCloudException(
+                "Error in updating user {user}: {message}".format(
+                    user=name_or_id, message=e.message))
+
+    def create_user(
+            self, name, password=None, email=None, project=None,
+            enabled=True):
+        """Create a user."""
+        try:
+            if project:
+                project_id = self._get_project(project).id
+            else:
+                project_id = None
+            self.keystone_client.users.create(
+                user_name=name, password=password, email=email,
+                project=project_id, enabled=enabled)
+        except Exception as e:
+            self.log.debug("keystone create user issue", exc_info=True)
+            raise OpenStackCloudException(
+                "Error in creating user {user}: {message}".format(
+                    user=name, message=e.message))
+
+    def delete_user(self, name_or_id):
+        try:
+            user = self._get_user(name_or_id)
+            self._user_manager.delete(user.id)
+        except Exception as e:
+            self.log.debug("keystone delete user issue", exc_info=True)
+            raise OpenStackCloudException(
+                "Error in deleting user {user}: {message}".format(
+                    user=name_or_id, message=e.message))
+
     def _get_glance_api_version(self):
         if 'image' in self.api_versions:
             return self.api_versions['image']
@@ -450,6 +652,52 @@ class OpenStackCloud(object):
                 return network
         return None
 
+    # TODO(Shrews): This will eventually need to support tenant ID and
+    # provider networks, which are admin-level params.
+    def create_network(self, name, shared=False, admin_state_up=True):
+        """Create a network.
+
+        :param name: Name of the network being created.
+        :param shared: Set the network as shared.
+        :param admin_state_up: Set the network administrative state to up.
+
+        :returns: The network object.
+        :raises: OpenStackCloudException on operation error.
+        """
+        neutron = self.neutron_client
+
+        network = {
+            'name': name,
+            'shared': shared,
+            'admin_state_up': admin_state_up
+        }
+
+        try:
+            net = neutron.create_network({'network': network})
+        except Exception as e:
+            self.log.debug("Network creation failed", exc_info=True)
+            raise OpenStackCloudException(
+                "Error in creating network %s: %s" % (name, e.message))
+        # Turns out neutron returns an actual dict, so no need for the
+        # use of meta.obj_to_dict() here (which would not work against
+        # a dict).
+        return net['network']
+
+    def delete_network(self, name_or_id):
+        """Delete a network.
+
+        :param name_or_id: Name or ID of the network being deleted.
+        :raises: OpenStackCloudException on operation error.
+        """
+        neutron = self.neutron_client
+        network = self.get_network(name_or_id)
+        try:
+            neutron.delete_network(network['id'])
+        except Exception as e:
+            self.log.debug("Network deletion failed", exc_info=True)
+            raise OpenStackCloudException(
+                "Error in deleting network %s: %s" % (name_or_id, e.message))
+
     def _get_images_from_cloud(self, filter_deleted):
         # First, try to actually get images from glance, it's more efficient
         images = dict()
@@ -489,7 +737,6 @@ class OpenStackCloud(object):
         image = self.get_image(image_id, exclude)
         if image:
             return image.id
-        self._image_cache[image_id] = None
         return None
 
     def get_image_id(self, image_name, exclude=None):
@@ -560,6 +807,7 @@ class OpenStackCloud(object):
     def _upload_image_v1(self, name, filename, **image_kwargs):
         image = self.glance_client.images.create(name=name, **image_kwargs)
         image.update(data=open(filename, 'rb'))
+        self._cache.invalidate()
         return self.get_image_dict(image.id)
 
     def _upload_image_v2(
@@ -1235,13 +1483,31 @@ class OpenStackCloud(object):
 
 class OperatorCloud(OpenStackCloud):
 
+    @property
+    def auth_token(self):
+        if self.auth_plugin in (None, "None", ''):
+            return self._auth_token
+        if not self._auth_token:
+            self._auth_token = self.keystone_session.get_token()
+        return self._auth_token
+
     @property
     def ironic_client(self):
         if self._ironic_client is None:
             ironic_logging = logging.getLogger('ironicclient')
             ironic_logging.addHandler(logging.NullHandler())
             token = self.auth_token
-            endpoint = self.get_endpoint(service_type='baremetal')
+            if self.auth_plugin in (None, "None", ''):
+                # TODO: This needs to be improved logic wise, perhaps a list,
+                # or enhancement of the data stuctures with-in the library
+                # to allow for things aside password authentication, or no
+                # authentication if so desired by the user.
+                #
+                # Attempt to utilize a pre-stored endpoint in the auth
+                # dict as the endpoint.
+                endpoint = self.auth['endpoint']
+            else:
+                endpoint = self.get_endpoint(service_type='baremetal')
             try:
                 self._ironic_client = ironic_client.Client(
                     '1', endpoint, token=token)
@@ -1319,3 +1585,53 @@ class OperatorCloud(OpenStackCloud):
                 "ironic machine unregistration failed", exc_info=True)
             raise OpenStackCloudException(
                 "Error unregistering machine from Ironic: %s" % e.message)
+
+    def validate_node(self, uuid):
+        try:
+            ifaces = self.ironic_client.node.validate(uuid)
+        except Exception as e:
+            self.log.debug(
+                "ironic node validation call failed", exc_info=True)
+            raise OpenStackCloudException(e.message)
+
+        if not ifaces.deploy or not ifaces.power:
+            raise OpenStackCloudException(
+                "ironic node %s failed to validate. "
+                "(deploy: %s, power: %s)" % (ifaces.deploy, ifaces.power))
+
+    def node_set_provision_state(self, uuid, state, configdrive=None):
+        try:
+            self.ironic_client.node.set_provision_state(
+                uuid,
+                state,
+                configdrive
+            )
+        except Exception as e:
+            self.log.debug(
+                "ironic node failed change provision state to %s" % state,
+                exc_info=True)
+            raise OpenStackCloudException(e.message)
+
+    def activate_node(self, uuid, configdrive=None):
+        self.node_set_provision_state(uuid, 'active', configdrive)
+
+    def deactivate_node(self, uuid):
+        self.node_set_provision_state(uuid, 'deleted')
+
+    def set_node_instance_info(self, uuid, patch):
+        try:
+            self.ironic_client.node.update(uuid, patch)
+        except Exception as e:
+            self.log.debug(
+                "Failed to update instance_info", exc_info=True)
+            raise OpenStackCloudException(e.message)
+
+    def purge_node_instance_info(self, uuid):
+        patch = []
+        patch.append({'op': 'remove', 'path': '/instance_info'})
+        try:
+            return self.ironic_client.node.update(uuid, patch)
+        except Exception as e:
+            self.log.debug(
+                "Failed to delete instance_info", exc_info=True)
+            raise OpenStackCloudException(e.message)
