@@ -52,7 +52,7 @@ def get_server_public_ip(server):
 def get_groups_from_server(cloud, server, server_vars):
     groups = []
 
-    region = cloud.region
+    region = cloud.region_name
     cloud_name = cloud.name
 
     # Create a group for the cloud
@@ -104,7 +104,7 @@ def get_hostvars_from_server(cloud, server, mounts=None):
     if interface_ip:
         server_vars['interface_ip'] = interface_ip
 
-    server_vars['region'] = cloud.region
+    server_vars['region'] = cloud.region_name
     server_vars['cloud'] = cloud.name
 
     flavor_id = server.flavor['id']
@@ -112,10 +112,15 @@ def get_hostvars_from_server(cloud, server, mounts=None):
     if flavor_name:
         server_vars['flavor']['name'] = flavor_name
 
-    image_id = server.image['id']
-    image_name = cloud.get_image_name(image_id)
-    if image_name:
-        server_vars['image']['name'] = image_name
+    # OpenStack can return image as a string when you've booted from volume
+    if unicode(server.image) == server.image:
+        image_id = server.image
+    else:
+        image_id = server.image.get('id', None)
+    if image_id:
+        image_name = cloud.get_image_name(image_id)
+        if image_name:
+            server_vars['image']['name'] = image_name
 
     volumes = []
     for vol in cloud.get_volumes(server):
